// Copyright 1994 by Jon Dart.  All Rights Reserved.

// Stand-alone executable to build the binary opening book from
// a text file.

// There are two book files, one for moves by White and one for
// Black.  Each book consists of three portions: a header, a hash
// table, and an array of book entries.

// The header is 8 bytes:
// offset
// 0 		version number
// 1-2		hash table size
// 3-4		number of book entries
// 5-7		unused

// The hash table is an array of unsigned indexes into the book
// entry array.  A hash code for the position modulo the hash table
// size is used to look up in this array the first book entry for the
// position.

// The book entries themselves are structures defined in bookentr.h.
// Each one is 6 bytes and contains the full hash code, the move
// index, and a weight indicating how often the computer should play
// the move.  Hash entries for each chain in the hash table are stored
// contiguously in the book, and the weight byte for the last entry
// in each chain has its hi bit set.

// Each book is limited to 65536 bytes, enough to hold slightly over
// 10,700 half-moves.

#include "stdafx.h"
#include "board.h"
#include "emove.h"
#include "bookentr.h"
// #include "bookwrit.h"
#include "hash.h"
#include "bhash.h"
#include "movegen.h"
#include "notation.h"
#include "types.h"
#include "debug.h"
extern "C"
{
#include <string.h>
#include <ctype.h>
};

#include <fstream.h>
#include <iostream.h>

const unsigned HASH_TABLE_SIZE = 511;
const unsigned MAX_MOVES = (65535 - HASH_TABLE_SIZE*2 - 8)/6;

struct book_info
{
    unsigned next_free;
    CWordArray *hash_table; 
    CPtrArray *book_moves; // array of Book_Entry *
};

// During opening book construction, we add an extra member to
// each book entry type.
class Book_Entry2  : public Book_Entry
{
public:
     Book_Entry2( unsigned long hc, byte rec, byte mv_indx,
	          unsigned int nxt, Boolean last = FALSE);

     unsigned int next;
};

Book_Entry2::Book_Entry2( unsigned long hc, byte rec, byte mv_indx,
            	          unsigned int nxt, Boolean last)
:Book_Entry(hc,rec,mv_indx,last),next(nxt)
{
}

static book_info book_data[2];

const int Entry_Size = 6; //sizeof(Book_Entry)
const int Header_Size = 8;
const int Buffer_Size = 4096;
const byte Book_Version = 3;

static Boolean
search(const unsigned start, const unsigned move_index,
       const ColorType side,
       const unsigned long hc)
{
   unsigned indx = start;
   Book_Entry2 be(hc, 0, move_index, 0, False);
   while (indx != INVALID)
   {
      _ASSERT(indx < MAX_MOVES);
      Book_Entry2 *entry = (Book_Entry2*)(*(book_data[side].book_moves))[indx];
      if (entry->is_equal(be) && entry->move_index == move_index)
	 return True;
      indx = entry->next;
   }
   return False;
}

static void
add_move(const Board & board, int move_index, int recommend)
{
   unsigned probe = (unsigned) (Board_Hash::HashCode2(board) % HASH_TABLE_SIZE);
   unsigned hit = (*(book_data[board.Side()].hash_table))[probe];
   if (hit == INVALID || !search(hit, move_index, board.Side(),
				 board.HashCode()))
   {
      Book_Entry2 *new_entry = new Book_Entry2(board.HashCode(), recommend,
					     move_index, hit, False);
      ASSERT(new_entry);
      if (book_data[board.Side()].next_free >= MAX_MOVES)
      {
         fprintf(stderr,"\nError - Too many moves in book!\n");
         exit(-1);
      }
      unsigned next = book_data[board.Side()].next_free;
      (*(book_data[board.Side()].book_moves))[next] = new_entry;
      (*(book_data[board.Side()].hash_table))[probe] = next;
#ifdef DEBUG
      cout << "p:" << probe << " h:" << (hex) << board.HashCode() << 
         (dec) << " i:"
	 << move_index << " f:" << next << " w:" << recommend << endl;
#endif
      book_data[board.Side()].next_free++;
   }
}

static void
write_book(const ColorType side)
{
   byte *write_buffer = new byte[Buffer_Size];
   ASSERT(write_buffer);
   static char *Book_File_Name[2] = {"bookb", "bookw"};
   ofstream book_file(Book_File_Name[side],
		       ios::out | ios::trunc | ios::binary);
   if (!book_file.good())
      return;

   // write the header

   book_file.put((char)Book_Version);
   unsigned ht_size = HASH_TABLE_SIZE;
   book_file.write((char *) &ht_size, (int) sizeof(uint16));
   book_file.write((char *) (&(book_data[side].next_free)),
		   (int) sizeof(uint16));
   // pad out the header to 8 bytes:
   char stuff[3] = {'\0', '\0', '\0'};
   book_file.write(stuff, 3);

   // Re-work the hash table.  Right now it holds pointers to the
   // head of each hash chain, linked together with "next" pointers.
   // We are going to rearrange the chains so they are contiguous
   // in the book.  Then the hash table will hold indexes to the start
   // of each block of hash entries - these indexes are counts of
   // how many Book_Entry objects lie before the start of the desired
   // block.

   int running_count = 0;
   for (unsigned i = 0; i < HASH_TABLE_SIZE; i++)
   {
      unsigned ht_entry = (*(book_data[side].hash_table))[i];
      if (ht_entry == INVALID)
      {
         book_file.put((char)(INVALID % 256));
         book_file.put((char)(INVALID / 256));
      }
      else
      {
         book_file.put((char)(running_count % 256));
         book_file.put((char)(running_count / 256));

         unsigned indx = ht_entry;
         unsigned count = 0;
         Book_Entry2 *entry = NULL;
         while (indx != INVALID)
         {
           entry = (Book_Entry2*)(*(book_data[side].book_moves))[indx];
           indx = entry->next;
           ++running_count;
         }
         if (entry)
           entry->set_last();
      }
   }

   // write the book moves

   unsigned index = 0;
   for (i = 0; i < HASH_TABLE_SIZE; i++)
   {
      unsigned ht_entry = (*(book_data[side].hash_table))[i];
      while (ht_entry != INVALID)
      {
         Book_Entry2 *book_entr = 
             (Book_Entry2*)(*(book_data[side].book_moves))[ht_entry];
         if (index + Entry_Size <= Buffer_Size)
         {
	    memcpy(write_buffer + index,
		   &book_entr->my_hash_code, Entry_Size);
	    index += Entry_Size;
         } 
         else
         {
	    unsigned to_go = Buffer_Size - index;
	    if (to_go)
	    {
	       memcpy(write_buffer + index,
	   	      &book_entr->my_hash_code, to_go);
            }
            book_file.write(write_buffer, Buffer_Size);
	    memcpy(write_buffer, (char*)(&book_entr->my_hash_code) + to_go, Entry_Size - to_go);
	    index = Entry_Size - to_go;
         }
         ht_entry = book_entr->next;
      }
   }
   if (index)
      book_file.write(write_buffer, index);
   book_file.close();
   delete[] write_buffer;

   cout << book_data[side].next_free << " moves in "
      << side << " book." << endl;
}

int
main(int argc, char **argv)
{
   char book_name[128];
   fstream infile;

   if (argc < 2)
      strcpy(book_name, "BOOK.TXT");
   else
      strcpy(book_name, argv[1]);

   infile.open(book_name, ios::in);
   if (!infile.good())
   {
      cerr << "Can't open book file: " << book_name << endl;
      return -1;
   }
   book_data[White].book_moves = new CPtrArray ();
   book_data[White].book_moves->SetSize(MAX_MOVES);
   book_data[Black].book_moves = new CPtrArray ();
   book_data[Black].book_moves->SetSize(MAX_MOVES);
   book_data[White].hash_table = new CWordArray ();
   book_data[White].hash_table->SetSize(HASH_TABLE_SIZE);
   book_data[Black].hash_table = new CWordArray ();
   book_data[Black].hash_table->SetSize(HASH_TABLE_SIZE);
   book_data[White].next_free = 0;
   book_data[Black].next_free = 0;

   char buf[128];
   char movebuf[20];
   unsigned line = 0;
   char *q;
   Board board;
   Board tmpboard;
   int recommend;
   Move moves[Move_Generator::MaxMoves];
   int dots = 0;

   for (unsigned i = 0; i < HASH_TABLE_SIZE; i++)
   {
      (*(book_data[White].hash_table))[i] = INVALID;
      (*(book_data[Black].hash_table))[i] = INVALID;
   }

   for (i = 0; i < MAX_MOVES; i++)
   {
      (*(book_data[White].book_moves))[i] = NULL;
      (*(book_data[Black].book_moves))[i] = NULL;
   }

   while (!infile.eof() && book_data[White].next_free < MAX_MOVES
	  && book_data[Black].next_free < MAX_MOVES)
   {
      recommend = 5;
      infile.getline(buf, 128);
#ifdef DEBUG
      cout << buf << endl;
#endif
      line++;
      char *p = buf;
      while (isspace(*p))
	 p++;
      switch (*p)
      {
      case '\0':
	 continue;
      case ';':
	 continue;
      case 'm':
	 tmpboard = board;
	 continue;
      case 'r':
	 board = tmpboard;
	 continue;
      case 's':
      case 't':
	 break;			// these are obsolete

      case '-':
	 board.Reset();
	 break;
      default:
	 {
	    while (*p)
	    {
	       while (isspace(*p))
		  p++;
	       if (*p == '\0')
	          break;
	       q = movebuf;
	       int count = 0;
	       while (!isspace(*p) && *p && count < 19)
	       {
		  *q++ = *p++;
		  ++count;
	       }
	       *q = '\0';
	       Move move = Notation::Value(board, board.Side(), movebuf);
	       if (move.IsNull())
	       {
		  cerr << endl << "Illegal move in file, line " << 
		     line << " (" << movebuf << ")" << endl;
		  infile.close();
		  return -1;
	       }
	       ExtendedMove emove(board, move);

	       // check (pseudo) legality:
	       Move_Generator mg(board, 0, Move::NullMove());
	       int found = 0;
	       int move_indx = 0;

	       int n = mg.Generate_Moves(moves, False, True);

	       for (int i = 0; i < n; i++)
	       {
		  if (moves[i] == emove)
		  {
		     move_indx = i;
		     found++;
		     break;
		  }
	       }
	       if (!found)
	       {
		  cerr << endl << "Illegal move in file, line " << line << " ("
		     << movebuf << ")" << endl;
		  infile.close();
		  return -1;
	       }
	       dots++;
	       if (dots % 50 == 0)
		  cout << '.' << (flush);
	       if (dots > 3000)
	       {
		  cout << endl;
		  dots = 1;
	       }
	       recommend = 5;
	       while (isspace(*p))
		  p++;
	       if (*p)
	       {
		  if ((*p >= '0') && (*p <= '9'))
		  {
		     recommend = *p - '0';
		     p++;
		  }
	       }
	       add_move(board, move_indx, recommend);
	       board.MakeMove(emove);
	    }
	 }
      }
   }
   infile.close();

   cout << endl;
   write_book(White);
   write_book(Black);
   return 0;
}
